import json
import os
import shutil


def generate_docs(collected_traces, directory='.'):
    package_dir = os.path.dirname(__file__)
    # convert sets to list to avoid JSON serialization error
    for trace_name in collected_traces:
        call_list = list(collected_traces[trace_name]['calls'])
        collected_traces[trace_name]['calls'] = call_list

    # get the leading/entry function
    numbered_traces = {v['call_number']: v for k, v in collected_traces.items()}
    keys = sorted(list(numbered_traces))
    for num in keys:
        if numbered_traces[num]['calls']:
            entry_function = numbered_traces[num]['function_name']
            break
    # add entry function to traces
    collected_traces['entry_function'] = entry_function

     # create order of functions // classes of docs html file
    order_of_functions = []

    def get_calls(trace_name):
        # print(trace_name)
        trace = collected_traces[trace_name]
        # print(trace['calls'])
        for call in trace['calls']:
            order_of_functions.append(collected_traces[call]['function_name'])
            get_calls(collected_traces[call]['function_name'])

    order_of_functions.append(entry_function)
    get_calls(entry_function)


    if not os.path.exists(directory+'/docs'):
        os.makedirs(directory+'/docs')

    # generate documentation in docs folder
    shutil.copy(package_dir+'/templates/index.html', directory+'/docs')
    shutil.copy(package_dir+'/templates/scroll.css', directory+'/docs')

    with open(directory+'/docs/traces_data.js', 'w') as tr:
        tr.write('let traces='+json.dumps(collected_traces))

    with open(directory+'/docs/functions_data.js', 'w') as fc:
        fc.write('let functions='+json.dumps(order_of_functions))



import collections

class OrderedSet(collections.OrderedDict, collections.MutableSet):
    # https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set

    def update(self, *args, **kwargs):
        if kwargs:
            raise TypeError("update() takes no keyword arguments")

        for s in args:
            for e in s:
                 self.add(e)

    def add(self, elem):
        self[elem] = None

    def discard(self, elem):
        self.pop(elem, None)

    def __le__(self, other):
        return all(e in other for e in self)

    def __lt__(self, other):
        return self <= other and self != other

    def __ge__(self, other):
        return all(e in self for e in other)

    def __gt__(self, other):
        return self >= other and self != other

    def __repr__(self):
        return 'OrderedSet([%s])' % (', '.join(map(repr, self.keys())))

    def __str__(self):
        return '{%s}' % (', '.join(map(repr, self.keys())))




"""SCROLL: Automatically generates documentation and unit-tests for python programs"""
__version__ = "2020.7.22"

import os
import sys
import click
import runpy
from scroll.runner import RuntimeContextManager
from scroll.tracer import Tracer
from scroll.generators import generate_docs
import pprint


@click.command()
@click.argument('filename')
def scroll(filename):
    """This is the main controller of the app.
        It aslo responds to commandline args using click
    """
    click.echo(f'SCROLL: Running file {filename}')
    # use a context manager to enter the file runtime and introspect it
    with RuntimeContextManager(Tracer) as manager:
        runpy.run_path(filename, run_name='__main__')
    sys.settrace(None)
    click.echo(f'SCROLL: Traces collected')
    collected_traces = manager.traces
    pprint.pprint(collected_traces)
    # generate docs and write to files
    directory = os.path.dirname(filename)
    generate_docs(collected_traces, directory)
    click.echo(f'SCROLL: Finished generating docs at docs/index.html')



import datetime

class Age:
    """
    test Ager DOCstring
    """

    def __init__(self):
        pass

    def get_age(self, yob):
        """
        Calculate age using date of birth.
            Args:
                yob: The year of birth.
            Returns:
                age: indicates how old the person is.
        """
        year_today = datetime.datetime.today().year
        age = year_today-yob
        return age


def calculate_bmi(weight, height):
    """
    Calculate body mass index.
        Args:
            wieght: in kilograms.
            height: in meters.
        Returns:
            bmi: body mass index.
    """
    bmi = weight/(height**2)
    return bmi


def main(yob, weight, height):
    """
    Generate the health credentials of a user based on input.
        Args:
            yob: The year of birth.
            wieght: in kilograms.
            height: in meters.
        Returns:
            age, bmi: as a dictionary.
    """
    ager = Age()
    age=ager.get_age(yob)
    bmi=calculate_bmi(weight, height)
    credentials = dict(
        age=age,
        bmi=bmi,
    )
    return credentials

creds = main(1993, 71, 1.79)



import inspect
import os
from pprint import pprint
from scroll.helpers import OrderedSet

class Tracer:
    def __init__(self):
        self.traces = {}
        self.call_number = 1
        self.cwd = os.getcwd()

    def __call__(self, frame, event, arg):
        code = frame.f_code
        func_name = code.co_name
        func_filename = code.co_filename

        # # Ignore write() calls from printing
        if func_name in ('write', '_shutdown'):
            return
        # # Ignore calls not in this module
        if 'venv' in func_filename \
            or func_filename.startswith('<') \
            or func_name.startswith('<') \
            or 'python' in func_filename:
            return

        # get function class name
        class_name = self.get_class_name(frame)
        if not 'None' in class_name:
            key_name = f"{class_name}.{func_name}"
        else:
            key_name = func_name

        if event == 'call':
            self.handle_call(frame, key_name, class_name)

        elif event == 'return':
            self.handle_return(arg, key_name, func_name, frame)
        return self

    def handle_call(self, frame, key_name, class_name):
        code = frame.f_code
        func_line_no = frame.f_lineno
        func_filename = code.co_filename

        arg_names = code.co_varnames[0:code.co_argcount]
        arg_dict = frame.f_locals

        # create a  dict of args and their types
        arg_values = {}
        for name in arg_names:
            if name in arg_dict:
                value = arg_dict[name]
                arg_values[name] = type(value).__name__
                # arg_values[name] = dict(
                #     arg=value,
                #     type=type(value)
                # )

        # called by?
        caller = frame.f_back
        if caller:
            caller_func = caller.f_code.co_name
            caller_line_no = caller.f_lineno
            caller_filename = caller.f_code.co_filename

            caller_class = self.get_class_name(caller)
            if not 'None' in caller_class:
                caller_key_name = f"{caller_class}.{caller_func}"
            else:
                caller_key_name = caller_func

            # record the call in the caller dict
            if self.traces.get(caller_key_name):
                # get the set of the functions called by the mother function and
                # add this current function
                callees = self.traces[caller_key_name]['calls']
                callees.add(key_name)
                self.traces[caller_key_name]['calls'] = callees

        else:
            pass

        self.traces[key_name] = dict(
            function_name=key_name,
            filename=func_filename.replace(self.cwd, ''),
            caller=caller_key_name,
            caller_line_num=caller_line_no,
            caller_file=caller_filename.replace(self.cwd, ''),
            line_num=func_line_no,
            call_args=arg_values,
            call_number=self.call_number,
            # docstring=docstring,
            class_name=class_name,
            calls=OrderedSet(),
        )

        self.call_number += 1

    def get_docstring(self, func_name, frame):
        if func_name == '__init__':
            class_obj = self.get_class_obj(frame)
            docstring = class_obj.__doc__
        else:
            # normal function
            func_details = frame.f_globals.get(func_name)
            docstring = func_details.__doc__
        return docstring

    def handle_return(self, arg, key_name, func_name, frame):
        # get docstring
        docstring = self.get_docstring(func_name, frame)
        if docstring is None:
            try:
                class_obj = self.get_class_obj(frame)
                class_method = class_obj.__dict__[func_name]
                docstring = class_method.__doc__
            except AttributeError:
                pass

        call_dict = self.traces.get(key_name, {})
        call_dict.update(
            dict(
                return_type=type(arg).__name__,
                docstring=docstring,
                # return_value=arg,
            )
        )
        self.traces[key_name] = call_dict

    def get_class_name(self, frame):
        class_obj = self.get_class_obj(frame)
        # print(dir(class_obj))
        if class_obj:
            class_name = class_obj.__name__
        else:
            class_name = 'None'
        return class_name

    def get_class_obj(self, frame):
        # initialize as None
        class_obj = None
        args, _, _, value_dict = inspect.getargvalues(frame)
        # we check the first parameter for the frame function is
        # named 'self'
        if len(args) and args[0] == 'self':
            # in that case, 'self' will be referenced in value_dict
            instance = value_dict.get('self', '')
            if instance:
                # return its class
                class_obj =  getattr(instance, '__class__', None)
        return class_obj



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['scroll']

package_data = \
{'': ['*'], 'scroll': ['templates/*']}

install_requires = \
['click']

entry_points = \
{'console_scripts': ['scroll = scroll:scroll']}

setup(name='scroll',
      version='2020.7.22',
      description='SCROLL: Automatically generates documentation and unit-tests for python programs',
      author='Edison Abahurire',
      author_email='abahedison@gmail.com',
      url='https://github.com/SimiCode/scroll',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



import sys
from scroll.tracer import Tracer
from scroll.runner import RuntimeContextManager
from demo.demo import main
import pprint
from scroll.generators import generate_docs


if __name__ == '__main__':
    with RuntimeContextManager(Tracer) as manager:
        creds = main(1993, 71, 1.79)
    # stop tracing
    sys.settrace(None)

    # do stuff with traces
    collected_traces = manager.traces
    pprint.pprint(collected_traces)
    generate_docs(collected_traces, directory='demo')



"""SCROLL: Automatically generates documentation and unit-tests for python programs"""
# this new verison sleeps for 30 seconds in case of any error/exception and kees retrying infinitely
'''
Main Entry file:
> Used for hooking into a running program using cli and/or a context manager
so that the program can be run as ..
scroll myprogram.py
'''
import sys


class RuntimeContextManager():
    def __init__(self, Tracer):
        self.tracer = Tracer()

    def __enter__(self):
        sys.settrace(self.tracer)
        return self.tracer

    def __exit__(self, type, value, traceback):
        pass
